unit contain;
{
        Container Lite  (C++ v 1.91)
        (C) Copyright 1996  John Webster Small
        All rights reserved
}

interface

uses sysutils;

const

     _CL_MAXNODES = 65520 div sizeof(pointer);
     CL_NOTFOUND : word  =  _CL_MAXNODES;
     CL_MAXNODES : word  =  _CL_MAXNODES;
     CL_LIMIT    : word  =  20;
     CL_DELTA    : word  =  10;
     CL_ALL_DESCENDANTS = TRUE;
     CL_ONLY_CLASS = FALSE;
     CL_FREE = TRUE;
     CL_BIND_ONLY = FALSE;

type

    EContainerException = class(Exception);

    CL_cmp = function
        (datum1, datum2 : TObject) : integer;
    CL_cmpKey = function
        (var key; datum : TObject) : integer;

    CL_LinkS_Array
      = array[0.._CL_MAXNODES-1] of TObject;
    CL_LinkS = ^CL_LinkS_Array;

    Container = class
      public
        constructor Create;
        constructor Create_specs(restrictTo : TClass;
            descendants : boolean;
            disposableNodes_ : boolean);
        destructor Destroy; override;
        function  limit : word;
        procedure setLimit(newLimit : word);
        procedure pack;
        function  delta : word;
        procedure setDelta(newDelta : word);
        function  nodes : word;
        function  maxNodes : word;
        function  vacancy : boolean;
        function  vacancyNonElastic : boolean;
        procedure setMaxNodes(newMaxNodes : word);
        function  typeAllowed(D : TObject) : boolean;
        function  rootType : TClass;
        function  polyCluster : boolean;
        function  atIns(n : word; D : TObject)
          : boolean;
        function  atInsFL(n : word; D : TObject)
          : boolean;
        function  insQ(D : TObject) : boolean;
        function  insQFL(D : TObject) : boolean;
        function  atPut(n : word; D : TObject)
          : boolean;
        function  atPutFL(n : word; D : TObject)
          : boolean;
        function  atXchg(n : word; D : TObject)
          : TObject;
        function  atGet(n : word) : TObject;
        property  cells[n : word] : TObject
          read atGet; default;
        function  atRmv(n : word) : TObject;
        procedure allRmv;
        procedure atFree(n : word);
        procedure allFree;
        procedure allClr;
        function  indexOf(D : TObject) : word;
        function  sorted : boolean;
        procedure unsort;
        procedure setcmp(cmp : CL_cmp);
        function  insSort(D : TObject) : boolean;
        function  insSortFL(D : TObject) : boolean;
        function  insUnique(D : TObject) : boolean;
        function  insUniqueFL(D : TObject) : boolean;
        procedure sort;
        function  matchFirst
          (D : TObject; var idx : word) : boolean;
        function  findFirst(var key;
            cmpKey : CL_cmpKey; var idx : word;
            honorSorted : boolean) : boolean;
      protected
        function  _attachD(D : TObject)
          : boolean; virtual;
        procedure _detachD(D : TObject); virtual;
      private
        _restrictTo : TClass;
        _descendants, _disposableNodes: boolean;
        lowLimit, lowThreshold, first : word;
        linkS : CL_LinkS;
        _limit, _delta, _nodes, _maxNodes : word;
        _sorted : boolean;
        _cmp : CL_cmp;
      public
        property disposableNodes : boolean
          read _disposableNodes write _disposableNodes;
    end;

implementation

constructor Container.Create;
begin
    Create_specs(TObject,true,true);
end;

constructor Container.Create_specs(restrictTo : TClass;
    descendants, disposableNodes_ : boolean);

begin
    inherited Create;
    _restrictTo := restrictTo;
    _descendants := descendants;
    _disposableNodes := disposableNodes_;
    _sorted := true;
    _cmp := nil;
    first := 0;
    _nodes := 0;
  {
        The following relationships are maintained
        during operation of a Container:
        1 <= _delta <= lowLimit <= _limit <= _maxNodes
                <= _CL_MAXNODES
        lowThreshold = lowLimit - _delta;
  }
    if (CL_MAXNODES > _CL_MAXNODES) then
        CL_MAXNODES := _CL_MAXNODES;
    if (CL_LIMIT > CL_MAXNODES) then
        CL_LIMIT := CL_MAXNODES;
    if (CL_DELTA > CL_LIMIT) then
        CL_DELTA := CL_LIMIT;
    if (CL_DELTA = 0) then begin
        CL_DELTA := 1;
        if (CL_LIMIT < CL_DELTA) then
            CL_LIMIT := CL_DELTA;
        if (CL_MAXNODES < CL_LIMIT) then
            CL_MAXNODES := CL_LIMIT;
        end;
    linkS := nil;
    _limit := CL_LIMIT;
    _delta := CL_DELTA;
    _maxNodes := CL_MAXNODES;
    lowLimit := _limit;
    lowThreshold := lowLimit - _delta;
end;

destructor Container.Destroy;
begin
    allClr;
    inherited Destroy;
end;

function  Container.limit : word;
  begin result := _limit; end;

procedure Container.setLimit(newLimit : word);
var newLinkS : CL_LinkS;
    newSize, headNodes : word;
begin
    if (newLimit < _nodes) then
        newLimit := _nodes
    else if (newLimit > _maxNodes) then
        newLimit := _maxNodes;
    if (newLimit < _delta) then
        newLimit := _delta;
    if ((newLimit = 0) or (newLimit = _limit)) then
        exit;
    if (linkS = nil) then begin
        _limit := newLimit;
        exit;
        end;
    newSize := sizeof(pointer) * newLimit;
    GetMem(newLinkS,newSize);
    if newLinkS = nil then
        exit;
    headNodes := _limit - first;
    if (headNodes > _nodes) then
        headNodes := _nodes;
    if (headNodes > 0) then  { move leading nodes }
        move(linkS^[first],newLinkS^,
            sizeof(pointer)*headNodes);
    { copy wrap around trailing nodes }
    if (headNodes < _nodes) then
        move(linkS^,newLinkS^[headNodes],
            sizeof(pointer)*(_nodes-headNodes));
    if (newLimit > _limit) then
        if (newLimit - _delta > _limit) then
            lowLimit := newLimit - _delta
        else
            lowLimit := _limit
    else
        if (newLimit - _delta > _delta) then
            lowLimit := newLimit - _delta
        else
            lowLimit := _delta;
    lowThreshold := lowLimit - _delta;
    FreeMem(linkS,sizeof(pointer)*_limit);
    linkS := newLinkS;
    _limit := newLimit;
    first := 0;
end;

procedure Container.pack;
  begin setLimit(_nodes); end;

function  Container.delta : word;
  begin result := _delta; end;

procedure Container.setDelta(newDelta : word);
begin
    if ((newDelta > 0) and (newDelta <= lowLimit))
        then _delta := newDelta;
end;

function  Container.nodes : word;
  begin result := _nodes; end;

function  Container.maxNodes : word;
  begin result := _maxNodes; end;

procedure Container.setMaxNodes(newMaxNodes : word);
begin
    if (newMaxNodes < _limit) then
        if (newMaxNodes > _nodes) then
            setLimit(newMaxNodes);
    if (newMaxNodes >= _limit) then
        if (newMaxNodes > _CL_MAXNODES) then
            _maxNodes := _CL_MAXNODES
        else
            _maxNodes := newMaxNodes;
end;

function  Container.vacancy : boolean;
begin
  result := _nodes < _maxNodes;
end;

function  Container.vacancyNonElastic : boolean;
begin
  result := _nodes < _limit;
end;

function  Container.typeAllowed(D : TObject) : boolean;
begin
    if (_descendants) then
        result := D is _restrictTo
    else
        result := D.ClassName = _restrictTo.ClassName;
end;

function  Container.rootType : TClass;
  begin  result := _restrictTo; end;

function  Container.polyCluster : boolean;
  begin result := _descendants; end;

function Container.atIns
    (n : word; D : TObject) : boolean;
var newLinkS : CL_LinkS;
    newSize, newLimit, i : word;
begin
    result := false;
    if (D = nil) then
        exit;
    if not typeAllowed(D) then
        raise EContainerException.Create
          ('wrong type for Container');
    if (_nodes = _limit)  then begin
        { no vacancy - expand }
        if (_limit = _maxNodes) then
            exit;
        if (_maxNodes - _delta > _limit) then
            newLimit := _limit + _delta
        else
            newLimit := _maxNodes;
        newSize := sizeof(pointer) * newLimit;
        GetMem(newLinkS,newSize);
        if newLinkS = nil then
            exit;  { expansion unavailable }
        if (not _attachD(D)) then begin
            { data refuses binding }
            FreeMem(newLinkS,newSize);
            exit;
        end;
        if (n > _nodes) then
            n := _nodes;
        newLinkS^[n] := D;
        { calc headNodes }
        i := _limit - first;
        if (n <= i) then begin
            { insert in head section }
            if (n > 0) then begin
                { postfix or split head }
                move(linkS^[first],newLinkS^,
                    sizeof(pointer)*n);
                if (n < i) then { it's a split }
                    move(linkS^[first+n],
                        newLinkS^[n+1],
                        sizeof(pointer)*(i-n));
            end
            else  { prefix head }
                move(linkS^[first],newLinkS^[1],
                    sizeof(pointer)*i);
            if (i < _nodes) then { copy tail }
                move(linkS^,newLinkS^[1+i],
                    sizeof(pointer) * (_nodes-i));
        end
        else begin  { insert in tail section }
            { copy head first }
            move(linkS^[first],newLinkS^,
                sizeof(pointer)*i);
            if (i < _nodes) then begin { copy tail }
                move(linkS^,newLinkS^[i],
                    sizeof(pointer)*(n-i));
                if (n < _nodes) then
                    { it's a tail split }
                    move(linkS^[n-i],
                        newLinkS^[n+1],
                        sizeof(pointer)
                        *(_nodes-n));
            end;
        end;
        { Compute next smaller linkS size
         and threshold for shrinking. }
        lowLimit := _limit;
        lowThreshold := lowLimit - _delta;
        { swap new for old }
        FreeMem(linkS,sizeof(pointer)*_limit);
        linkS := newLinkS;
        _limit := newLimit;
        first := 0;
    end
    else begin  { vacancy }
        if (linkS = nil) then begin
            newSize := sizeof(pointer) * _limit;
            GetMem(linkS,newSize);
            if linkS = nil then
                exit;
            end;
        if (not _attachD(D)) then begin
            if (_nodes = 0) then begin
                FreeMem(linkS,sizeof(pointer)*_limit);
                linkS := nil;
            end;
            exit;
        end;
        if (n = 0) then begin { push }
            if first > 0 then
                dec(first)
            else
                first := _limit - 1;
            linkS^[first] := D;
        end
        else if (n >= _nodes) then begin { insQ }
            n := _nodes;
            linkS^[(first+n) mod _limit] := D;
            end
        else begin { insert interior }
            i := (first + n) mod _limit;
            if ((i < first) or (first = 0)) then
                { move rear rightward }
                move(links^[i],linkS^[i+1],
                    sizeof(pointer) * (_nodes-n))
            else begin { move front leftward }
                move(links^[first],
                    linkS^[first-1],
                    sizeof(pointer)*n);
                dec(first);
                dec(i);
            end;
            linkS^[i] := D;
        end;
    end;
    inc(_nodes);
    _sorted := false;
    result := true;
end;

function  Container.atInsFL
    (n : word; D : TObject) : boolean;
var ok : boolean;
begin
    if not _disposableNodes then
       raise EContainerException.Create
         ('vector does not own dynamic node');
    ok := atIns(n,D);
    if (not ok) and (D <> nil) then
       D.Free;
    result := ok;
end;

function Container.insQ(D : TObject) : boolean;
  begin result := atIns(nodes,D); end;

function Container.insQFL(D : TObject) : boolean;
  begin result := atInsFL(nodes,D); end;

function Container.atPut
    (n : word; D : TObject) : boolean;
var oldDptr : ^TObject;
begin
    result := false;
    if (D = nil) then
        exit;
    if not typeAllowed(D) then
        raise EContainerException.Create
          ('wrong type for Container');
    if ((linkS <> nil) and (n < _nodes))
        then begin
        oldDptr := @linkS^[(first+n) mod _limit];
        if (D <> oldDptr^) then if (_attachD(D))
            then begin
            _detachD(oldDptr^);
            if _disposableNodes then
               oldDptr^.Free;
            result := true;
            oldDptr^ := D;
            _sorted := false;
            end;
        end;
end;

function  Container.atPutFL
    (n : word; D : TObject) : boolean;
var ok : boolean;
begin
    if not _disposableNodes then
       raise EContainerException.Create
       ('vector does not own dynamic node');
    ok := atPut(n,D);
    if (not ok) and (D <> nil) then
       D.Free;
    result := ok;
end;

function Container.atXchg
    (n : word; D : TObject) : TObject;
var oldDptr : ^TObject;
begin
    result := nil;
    if ((linkS <> nil) and (D <> nil) and (n < _nodes))
        then if (_attachD(D)) then begin
        _sorted := false;
        oldDptr := @(linkS^[(first+n) mod _limit]);
        result := oldDptr^;
        _detachD(oldDptr^);
        oldDptr^ := D;
        end;
end;

function  Container.atGet(n : word) : TObject;
begin
    if ((linkS <> nil) and (n < _nodes)) then
        result := linkS^[(first+n) mod _limit]
    else
        result := nil;
end;

function  Container.atRmv(n : word) : TObject;
var newLinkS : CL_LinkS;
    newSize, newLimit, i : word;
    D : TObject;
begin
    newLinkS := nil;
    if ((linkS = nil) or (n >= _nodes)) then begin
        result := nil;
        exit;
        end;
    i := (first+n) mod _limit;
    D := linkS^[i];
    if (_nodes <= lowThreshold) then begin
        { get ready to contract }
        newLimit := lowLimit;
        newSize := sizeof(pointer) * newLimit;
        GetMem(newLinkS,newSize);
        { Compute next smaller linkS
         size and threshold for
         shrinking. }
        if (newLinkS <> nil) then begin
            if (_nodes < _delta) then
                lowThreshold := 0
            else
                dec(lowThreshold,_delta);
            lowLimit := lowThreshold + _delta;
            end;
    end;
    if (newLinkS <> nil) then begin
        { remove and contract }
        { calc head nodes }
        i := _limit - first;
        if i > _nodes then
           i := _nodes;
        if (n < i) then begin
            { rmv from head section }
            if (n > 0) then begin { rmv from head }
                move(linkS^[first],newLinkS^,
                    sizeof(pointer)*n);
                if (n < i - 1) then
                    move(linkS^[first+n+1],
                        newLinkS^[n],
                        sizeof(pointer)*(i-n-1));
            end
            else if (i > 1)  then { pop head }
                move(linkS^[first+1],newLinkS^,
                    sizeof(pointer)*(i-1));
            if (i < _nodes) then { copy tail }
                move(linkS^,newLinkS^[i-1],
                    sizeof(pointer)*(_nodes-i));
        end
        else begin  { i <= n < nodes }
            { rmv from tail section
              copy head first }
            move(linkS^[first],newLinkS^,
                sizeof(pointer)*i);
            if (i = n) then begin { pop tail }
                if (n < _nodes - 1) then
                    { copy trailing tail }
                    move(linkS^[1],newLinkS^[i],
                        sizeof(pointer)*(_nodes-n-1));
            end
            else begin { copy up to n in tail }
                move(linkS^,newLinkS^[i],
                    sizeof(pointer)*(n-i));
                { skip n'th node }
                if (n < _nodes - 1) then
                   { copy anything trailing n'th node }
                    move(linkS^[n-i+1],newLinkS^[n],
                        sizeof(pointer)*(_nodes-n-1));
            end
        end;
        { swap new for old }
        FreeMem(linkS,sizeof(pointer)*_limit);
        linkS := newLinkS;
        _limit := newLimit;
        first := 0;
    end
    else begin { simply remove }
        if (n = 0) then begin  {  pop }
            inc(first);
            if (first >= _limit) then
                first := 0;
        end
        else if (n <> _nodes-1) then { del interior }
            if (i < first) then { move tail leftward }
                move(linkS^[i+1],linkS^[i],
                    sizeof(pointer)*(_nodes-n-1))
            else begin { move head rightward }
                move(linkS^[first],linkS^[first+1],
                    sizeof(pointer)*n);
                inc(first);
            end
    end;
    dec(_nodes);
    if (_nodes = 0) then begin
        FreeMem(linkS,sizeof(pointer)*_limit);
        linkS := nil;
        _sorted := true;
    end;
    _detachD(D);
    result := D;
end;

procedure Container.allRmv;
var i, j : word;
begin
    if (linkS <> nil) then begin
        i := first;
        j := _nodes;
        while (j > 0) and (i < _limit) do begin
            _detachD(linkS^[i]);
            inc(i);
            dec(j);
            end;
        i := 0;
        while (j > 0) do begin
            _detachD(linkS^[i]);
            inc(i);
            dec(j);
            end;
        FreeMem(linkS,sizeof(pointer)*_limit);
        _sorted := true;
        first := 0;
        _nodes := 0;
        linkS := nil;
        _limit := _delta;
        lowLimit := _limit;
        lowThreshold := lowLimit - _delta;
        end;
end;

procedure Container.atFree(n : word);
var D : TObject;
begin
    if not _disposableNodes then
       raise EContainerException.Create
         ('vector does not own dynamic node')
    else begin
        D := atRmv(n);
        if D <> nil then
            D.Free;
        end;
end;

procedure Container.allFree;
var i, j : word;
begin
    if not _disposableNodes then
       raise EContainerException.Create
         ('vector does not own dynamic node')
    else if (linkS <> nil) then begin
        i := first;
        j := _nodes;
        while (j > 0) and (i < _limit) do begin
            _detachD(linkS^[i]);
            (linkS^[i]).Free;
            inc(i);
            dec(j);
            end;
        i := 0;
        while (j > 0) do begin
            _detachD(linkS^[i]);
            (linkS^[i]).Free;
            inc(i);
            dec(j);
            end;
        FreeMem(linkS,sizeof(pointer)*_limit);
        _sorted := true;
        first := 0;
        _nodes := 0;
        linkS := nil;
        _limit := _delta;
        lowLimit := _limit;
        lowThreshold := lowLimit - _delta;
        end;
end;

procedure Container.allClr;
begin
    if _disposableNodes then
        allFree
    else
        allRmv;
end;

function  Container.indexOf(D : TObject) : word;
var i, j : word;
begin
    if (linkS <> nil) then begin
        i := first;
        j := _nodes;
        while (j > 0) and (i < _limit) do begin
            if linkS^[i] = D then begin
               result := i;
               exit;
               end;
            inc(i);
            dec(j);
            end;
        i := 0;
        while (j > 0) do begin
            if linkS^[i] = D then begin
               result := i;
               exit;
               end;
            inc(i);
            dec(j);
            end;
        end;
    result := CL_NOTFOUND;
end;

function  Container.sorted : boolean;
    begin result := _sorted; end;

procedure Container.unsort;
    begin _sorted := false; end;

procedure Container.setcmp(cmp : CL_cmp);
begin
    _cmp := cmp;
    if _nodes > 0 then
        _sorted := false;
end;

function  Container.insSort(D : TObject) : boolean;
var low, mid, high : word;
    ok : boolean;
begin
    result := false;
    if (not Assigned(_cmp)) or (D = nil) then
        exit;
    if not typeAllowed(D) then
        raise EContainerException.Create
          ('wrong type for Container');
    if not _sorted then begin
        sort;
        if not _sorted then
            exit;
        end;
    low := 0;
    high := _nodes;  { linkS == 0 -> nodes == 0! }
    while (low < high)  do begin
        mid := ((high + low) shr 1);
        if _cmp(D,linkS^[(first+mid)
           mod _limit]) <= 0 then
          high := mid
        else
          low := mid + 1;
    end;
    ok := atIns(high,D);
    _sorted := true;
    result := true;
end;

function  Container.insSortFL(D : TObject) : boolean;
var ok : boolean;
begin
    if not _disposableNodes then
       raise EContainerException.Create
         ('vector does not own dynamic node');
    ok := insSort(D);
    if (not ok) and (D <> nil) then
        D.Free;
    result := ok;
end;

function  Container.insUnique(D : TObject) : boolean;
var idx : word;
begin
    if matchFirst(D,idx) then
        result := false
    else begin
        if not _sorted then
            result := insQ(D)
        else
            result := insSort(D);
    end;
end;

function  Container.insUniqueFL(D : TObject) : boolean;
var idx : word;
begin
    if not _disposableNodes then
       raise EContainerException.Create
         ('vector does not own dynamic node');
    if matchFirst(D,idx) then begin
        result := false;
        if D <> nil then
            D.Free;
        end
    else begin
        if not _sorted then
            result := insQFL(D)
        else
            result := insSortFL(D);
    end;
end;

procedure Container.sort;
var low, mid, high, d : word;
    D_ : TObject;
begin
    if not Assigned(_cmp)
        then exit;
    _sorted := true;
    if _nodes = 0 then
        exit;
    if (first > 0)  then begin
        { form contiguous block at front }
        d := (first + _nodes) mod _limit;
        if (d > first) then  { no wrap }
            move(linkS^[first],linkS^,
                sizeof(pointer)*_nodes)
        else if (d < first) then
            move(linkS^[first],linkS^[d],
                sizeof(pointer)*(_limit-first));
        { else array is full/contiguous }
        first := 0;
    end;
    high := 1; d := 1;
    while d < _nodes do begin
        low := 0;
        D_ := linkS^[d];
        while (low < high)  do begin
            mid := ((high + low) shr 1);
            if _cmp(D_,linkS^[mid])
                <= 0 then
                high := mid
            else
                low := mid + 1;
        end;
        if (high < d) then begin
            move(linkS^[high],linkS^[high+1],
                sizeof(pointer)*(d-high));
            linkS^[high] := D_;
        end;
        inc(d);
        high := d;
    end;
end;

function  Container.matchFirst
    (D : TObject; var idx : word) : boolean;
var low, high : word;
begin
    if (not Assigned(_cmp)) or (D = nil) then begin
        result := false;
        exit;
        end;
    if _sorted  then begin
        low := 0; high := _nodes - 1;
        { binary search for first match }
        while (low < high) do begin
            idx := ((high + low) shr 1);
            if _cmp(D,linkS^[(first+idx)
                mod _limit]) <= 0 then
                high := idx
            else
                low := idx + 1;
        end;
        idx := high;
        if _cmp(D,linkS^[(first+high)
            mod _limit]) <> 0 then
            idx := _nodes;
    end
    else  begin { linear search! }
        idx := 0;
        repeat
            if _cmp(D,linkS^[(first+idx)
                mod _limit]) = 0 then
                break;
            inc(idx);
        until idx >= _nodes;
    end;
    result := (idx < _nodes);
end;

function  Container.findFirst(var key;
    cmpKey : CL_cmpKey; var idx : word;
    honorSorted : boolean) : boolean;
var low, high : word;
begin
    if honorSorted and _sorted  then begin
        low := 0; high := _nodes - 1;
        { binary search for first match }
        while (low < high) do begin
            idx := ((high + low) shr 1);
            if cmpKey(key,linkS^[(first+idx)
                mod _limit]) <= 0 then
                high := idx
            else
                low := idx + 1;
        end;
        idx := high;
        if cmpKey(key,linkS^[(first+high)
            mod _limit]) <> 0 then
            idx := _nodes;
    end
    else  begin { linear search! }
        idx := 0;
        repeat
            if cmpKey(key,linkS^[(first+idx)
                mod _limit]) = 0 then
                break;
            inc(idx);
        until idx >= _nodes;
    end;
    result := (idx < _nodes);
end;

function  Container._attachD(D : TObject) : boolean;
  begin result := true; end;

procedure Container._detachD(D : TObject);
  begin end;

end.
